"""
LanceDB similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.storage as fos
import fiftyone.core.utils as fou
import fiftyone.brain.internal.core.utils as fbu
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)

lancedb = fou.lazy_import("lancedb")
pa = fou.lazy_import("pyarrow")


_SUPPORTED_METRICS = {
    "cosine": "cosine",
    "euclidean": "l2",
}

logger = logging.getLogger(__name__)


class LanceDBSimilarityConfig(SimilarityConfig):
    """Configuration for a LanceDB similarity instance.

    Args:
        table_name (None): the name of the LanceDB table to use. If none is
            provided, a new table will be created
        metric ("cosine"): the embedding distance metric to use when creating a
            new index. Supported values are ``("cosine", "euclidean")``
        uri ("/tmp/lancedb"): the database URI to use
        **kwargs: keyword arguments for :class:`SimilarityConfig`
    """

    def __init__(
        self,
        table_name=None,
        metric="cosine",
        uri="/tmp/lancedb",
        **kwargs,
    ):
        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.table_name = table_name
        self.metric = metric

        # store privately so these aren't serialized
        self._uri = uri

    @property
    def method(self):
        return "lancedb"

    @property
    def uri(self):
        return self._uri

    @uri.setter
    def uri(self, value):
        self._uri = value

    @property
    def max_k(self):
        return None

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(self, uri=None):
        self._load_parameters(uri=uri)


class LanceDBSimilarity(Similarity):
    """LanceDB similarity factory.

    Args:
        config: a :class:`LanceDBSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("lancedb")

    def ensure_usage_requirements(self):
        fou.ensure_package("lancedb")

    def initialize(self, samples, brain_key):
        return LanceDBSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class LanceDBSimilarityIndex(SimilarityIndex):
    """Class for interacting with LanceDB similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`LanceDBSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`LanceDBSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._table = None
        self._db = None
        self._initialize()

    def _initialize(self):
        try:
            db = lancedb.connect(self.config.uri)
        except Exception as e:
            raise ValueError(
                "Failed to connect to LanceDB backend at URI '%s'. Refer to "
                "https://docs.voxel51.com/integrations/lancedb.html for more "
                "information" % self.config.uri
            ) from e

        table_names = db.table_names()

        if self.config.table_name is None:
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            table_name = fbu.get_unique_name(root, table_names)

            self.config.table_name = table_name
            self.save_config()

        if self.config.table_name in table_names:
            table = db.open_table(self.config.table_name)
        else:
            table = None

        self._db = db
        self._table = table

    @property
    def table(self):
        """The ``lancedb.LanceTable`` instance for this index."""
        return self._table

    @property
    def total_index_size(self):
        if self._table is None:
            return 0

        return len(self._table)

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
    ):
        if self._table is None:
            pa_table = pa.Table.from_arrays(
                [[], [], []], names=["id", "sample_id", "vector"]
            )
        else:
            pa_table = self._table.to_arrow()

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            existing_ids = set(pa_table["id"].to_pylist()) & set(ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)

        if label_ids is not None:
            ids = list(label_ids)
        else:
            ids = list(sample_ids)

        dim = embeddings.shape[1]

        if self._table:
            prev_embeddings = np.concatenate(
                pa_table["vector"].to_numpy()
            ).reshape(-1, dim)
            embeddings = np.concatenate([prev_embeddings, embeddings])
            ids = pa_table["id"].to_pylist() + ids
            sample_ids = pa_table["sample_id"].to_pylist() + sample_ids

        embeddings = pa.array(embeddings.reshape(-1), type=pa.float32())
        embeddings = pa.FixedSizeListArray.from_arrays(embeddings, dim)
        sample_ids = list(sample_ids)
        pa_table = pa.Table.from_arrays(
            [ids, sample_ids, embeddings], names=["id", "sample_id", "vector"]
        )
        self._table = self._db.create_table(
            self.config.table_name, pa_table, mode="overwrite"
        )

        if reload:
            self.reload()

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = list(self._index.fetch(ids).vectors.keys())
            missing_ids = set(ids) - set(existing_ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, next(iter(missing_ids)))
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

                ids = existing_ids

        df = self._table.to_pandas()
        df = df[~df["id"].isin(ids)]
        self._table = self._db.create_table(
            self.config.table_name, df, mode="overwrite"
        )

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        df = self._table.to_pandas()

        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []
        missing_ids = []

        if sample_ids is not None and self.config.patches_field is not None:
            df.set_index("sample_id", drop=False, inplace=True)

            if not etau.is_container(sample_ids):
                sample_ids = [sample_ids]

            for sample_id in sample_ids:
                if sample_id in df.index:
                    found_embeddings.append(df.loc[sample_id]["vector"])
                    found_sample_ids.append(sample_id)
                    found_label_ids.append(df.loc[sample_id]["id"])
                else:
                    missing_ids.append(sample_id)
        elif self.config.patches_field is not None:
            df.set_index("id", drop=False, inplace=True)

            if label_ids is None:
                label_ids = list(df.index)
            elif not etau.is_container(label_ids):
                label_ids = [label_ids]

            for label_id in label_ids:
                if label_id in df.index:
                    found_embeddings.append(df.loc[label_id]["vector"])
                    found_sample_ids.append(df.loc[label_id]["sample_id"])
                    found_label_ids.append(label_id)
                else:
                    missing_ids.append(label_id)
        else:
            df.set_index("id", drop=False, inplace=True)

            if sample_ids is None:
                sample_ids = list(df.index)
            elif not etau.is_container(sample_ids):
                sample_ids = [sample_ids]

            for sample_id in sample_ids:
                if sample_id in df.index:
                    found_embeddings.append(df.loc[sample_id]["vector"])
                    found_sample_ids.append(sample_id)
                else:
                    missing_ids.append(sample_id)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(found_embeddings)
        sample_ids = np.array(found_sample_ids)
        if label_ids is not None:
            label_ids = np.array(found_label_ids)

        return embeddings, sample_ids, label_ids

    def cleanup(self):
        if self._db is None:
            return

        for tbl in (
            self.config.table_name,
            self.config.table_name + "_filter",
        ):
            if tbl in self._db.table_names():
                self._db.drop_table(tbl)

        self._table = None

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        if query is None:
            raise ValueError("LanceDB does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "LanceDB does not support least similarity queries"
            )

        if aggregation not in (None, "mean"):
            raise ValueError(
                f"LanceDB does not support {aggregation} aggregation"
            )

        if k is None:
            k = self.index_size

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        table = self._table

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = list(self.current_label_ids)
            else:
                index_ids = list(self.current_sample_ids)

            df = table.to_pandas()
            df.set_index("id", drop=False, inplace=True)
            df = df.loc[index_ids]
            table = self._db.create_table(
                self.config.table_name + "_filter", df, mode="overwrite"
            )

        metric = _SUPPORTED_METRICS[self.config.metric]

        ids = []
        dists = []
        for q in query:
            results = table.search(q).metric(metric).limit(k).to_df()
            ids.append(results.id.tolist())
            if return_dists:
                dists.append(results._distance.tolist())

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        df = self._table.to_pandas()
        df.set_index("id", drop=False, inplace=True)
        df = df.loc[query_ids]
        query = np.array([v for v in df["vector"]])
        if single_query:
            query = query[0, :]

        return query

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
Brain interface tests.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import unittest

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz


def test_uniqueness():
    dataset = foz.load_zoo_dataset("quickstart").clone()

    fob.compute_uniqueness(dataset)
    print(dataset.list_brain_runs())
    print(dataset.get_brain_info("uniqueness"))
    print(dataset.bounds("uniqueness"))

    dataset.delete_brain_runs()
    print(dataset)


def test_detection_mistakenness():
    dataset = foz.load_zoo_dataset("quickstart").clone()

    fob.compute_mistakenness(
        dataset, "predictions", label_field="ground_truth", copy_missing=True
    )
    print(dataset.list_brain_runs())
    print(dataset.get_brain_info("mistakenness"))

    # should be non-trivial
    print(dataset.bounds("mistakenness"))
    print(dataset.bounds("possible_missing"))
    print(dataset.bounds("possible_spurious"))
    print(dataset.bounds("ground_truth.detections.mistakenness"))
    print(dataset.bounds("ground_truth.detections.mistakenness_loc"))
    print(dataset.count_values("ground_truth.detections.possible_spurious"))
    print(dataset.count_values("predictions.detections.possible_missing"))
    print(dataset.count_values("ground_truth.detections.possible_missing"))

    dataset.delete_brain_runs()
    print(dataset)

    # should be None
    print(dataset.bounds("ground_truth.detections.mistakenness"))
    print(dataset.bounds("ground_truth.detections.mistakenness_loc"))
    print(dataset.count_values("ground_truth.detections.possible_spurious"))
    print(dataset.count_values("predictions.detections.possible_missing"))
    print(dataset.count_values("ground_truth.detections.possible_missing"))


def test_classification_mistakenness_confidence():
    dataset = foz.load_zoo_dataset("quickstart").clone()
    test_view = dataset.take(10)

    # labels proxy
    model = foz.load_zoo_model("alexnet-imagenet-torch")
    test_view.apply_model(model, "alexnet")

    # predictions proxy
    model = foz.load_zoo_model("resnet50-imagenet-torch")
    test_view.apply_model(model, "resnet50")

    fob.compute_mistakenness(test_view, "resnet50", label_field="alexnet")
    print(dataset.list_brain_runs())
    print(dataset.load_brain_view("mistakenness"))
    print(dataset.bounds("mistakenness"))

    dataset.delete_brain_runs()
    print(dataset)


def test_classification_mistakenness_logits():
    dataset = foz.load_zoo_dataset("quickstart").clone()
    test_view = dataset.take(10)

    # labels proxy
    model = foz.load_zoo_model("alexnet-imagenet-torch")
    test_view.apply_model(model, "alexnet")

    # predictions proxy
    model = foz.load_zoo_model("resnet50-imagenet-torch")
    test_view.apply_model(model, "resnet50", store_logits=True)

    fob.compute_mistakenness(
        test_view, "resnet50", label_field="alexnet", use_logits=True
    )
    print(dataset.list_brain_runs())
    print(dataset.load_brain_view("mistakenness"))
    print(dataset.bounds("mistakenness"))

    dataset.delete_brain_runs()
    print(dataset)


def test_hardness():
    dataset = foz.load_zoo_dataset("quickstart").clone()
    test_view = dataset.take(10)
    model = foz.load_zoo_model("alexnet-imagenet-torch")
    test_view.apply_model(model, "alexnet", store_logits=True)

    fob.compute_hardness(test_view, "alexnet")
    print(dataset.list_brain_runs())
    print(dataset.get_brain_info("hardness"))
    print(dataset.load_brain_view("hardness"))
    print(dataset.bounds("hardness"))

    dataset.delete_brain_runs()
    print(dataset)


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Mistakenness methods.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging
from math import exp

import numpy as np
from scipy.special import softmax
from scipy.stats import entropy

from fiftyone import ViewField as F
import fiftyone.core.brain as fob
import fiftyone.core.labels as fol
import fiftyone.core.media as fom
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov


logger = logging.getLogger(__name__)


_ALLOWED_TYPES = (
    fol.Classification,
    fol.Classifications,
    fol.Detections,
    fol.Polylines,
    fol.Keypoints,
    fol.TemporalDetections,
)
_MISSED_CONFIDENCE_THRESHOLD = 0.95
_DETECTION_IOU = 0.5


def compute_mistakenness(
    samples,
    pred_field,
    label_field,
    mistakenness_field,
    missing_field,
    spurious_field,
    use_logits,
    copy_missing,
    progress,
):
    """See ``fiftyone/brain/__init__.py``."""

    #
    # Algorithm
    #
    # The chance of a mistake is related to how confident the model prediction
    # was as well as whether or not the prediction is correct. A prediction
    # that is highly confident and incorrect is likely to be a mistake. A
    # prediction that is low confidence and incorrect is not likely to be a
    # mistake.
    #
    # Let us compute a confidence measure based on negative entropy of logits:
    # $c = -entropy(logits)$. This value is large when there is low uncertainty
    # and small when there is high uncertainty. Let us define modulator, $m$,
    # based on whether or not the answer is correct. $m = -1$ when the label is
    # correct and $1$ otherwise. Then, mistakenness is computed as
    # $(m * exp(c) + 1) / 2$ so that high confidence correct predictions result
    # in low mistakenness, high confidence incorrect predictions result in high
    # mistakenness, and low confidence predictions result in middling
    # mistakenness.
    #

    fov.validate_collection_label_fields(
        samples, (pred_field, label_field), _ALLOWED_TYPES, same_type=True
    )

    if samples.media_type == fom.VIDEO:
        mistakenness_field, _ = samples._handle_frame_field(mistakenness_field)
        missing_field, _ = samples._handle_frame_field(missing_field)
        spurious_field, _ = samples._handle_frame_field(spurious_field)

    is_objects = samples._is_label_field(
        pred_field,
        (fol.Detections, fol.Polylines, fol.Keypoints, fol.TemporalDetections),
    )
    if is_objects:
        eval_key = _make_eval_key(samples, mistakenness_field)
        config = DetectionMistakennessConfig(
            pred_field,
            label_field,
            mistakenness_field,
            missing_field,
            spurious_field,
            use_logits,
            copy_missing,
            eval_key,
        )
    else:
        eval_key = None
        config = ClassificationMistakennessConfig(
            pred_field, label_field, mistakenness_field, use_logits
        )

    brain_key = mistakenness_field
    brain_method = config.build()
    brain_method.ensure_requirements()
    brain_method.register_run(samples, brain_key, cleanup=False)
    brain_method.register_samples(samples)

    if is_objects:
        samples.evaluate_detections(
            pred_field,
            gt_field=label_field,
            eval_key=eval_key,
            classwise=False,
            iou=_DETECTION_IOU,
            progress=progress,
        )

    view = samples.select_fields([label_field, pred_field])
    processing_frames = samples._is_frame_field(label_field)

    logger.info("Computing mistakenness...")
    for sample in view.iter_samples(progress=progress):
        if processing_frames:
            images = sample.frames.values()
        else:
            images = [sample]

        sample_mistakenness = []
        num_missing = 0
        num_spurious = 0
        for image in images:
            if is_objects:
                (
                    img_mistakenness,
                    img_missing,
                    img_spurious,
                ) = brain_method.process_image(image, eval_key)

                num_missing += img_missing
                num_spurious += img_spurious
                if processing_frames:
                    image[missing_field] = img_missing
                    image[spurious_field] = img_spurious
            else:
                img_mistakenness = brain_method.process_image(image)

            if img_mistakenness is not None:
                sample_mistakenness.append(img_mistakenness)

            if processing_frames:
                image[mistakenness_field] = img_mistakenness

        if sample_mistakenness:
            sample[mistakenness_field] = np.max(sample_mistakenness)
        else:
            sample[mistakenness_field] = None

        if is_objects:
            sample[missing_field] = num_missing
            sample[spurious_field] = num_spurious

        sample.save()

    if eval_key is not None:
        samples.delete_evaluation(eval_key)

    brain_method.save_run_results(samples, brain_key, None)

    logger.info("Mistakenness computation complete")


# @todo move to `fiftyone/brain/mistakenness.py`
# Don't do this hastily; `get_brain_info()` on existing datasets has this
# class's full path in it and may need migration
class MistakennessMethodConfig(fob.BrainMethodConfig):
    def __init__(self, pred_field, label_field, mistakenness_field, **kwargs):
        super().__init__(**kwargs)
        self.pred_field = pred_field
        self.label_field = label_field
        self.mistakenness_field = mistakenness_field

    @property
    def type(self):
        return "mistakenness"


class MistakennessMethod(fob.BrainMethod):
    def __init__(self, config):
        super().__init__(config)
        self.pred_field = None
        self.label_field = None
        self.label_type = None

    def ensure_requirements(self):
        pass

    def register_samples(self, samples):
        self.pred_field, _ = samples._handle_frame_field(
            self.config.pred_field
        )
        self.label_field, _ = samples._handle_frame_field(
            self.config.label_field
        )
        self.label_type = samples._get_label_field_type(self.config.pred_field)

    def _validate_run(self, samples, brain_key, existing_info):
        self._validate_fields_match(brain_key, "pred_field", existing_info)
        self._validate_fields_match(brain_key, "label_field", existing_info)
        self._validate_fields_match(
            brain_key, "mistakenness_field", existing_info
        )


# @todo move to `fiftyone/brain/mistakenness.py`
# Don't do this hastily; `get_brain_info()` on existing datasets has this
# class's full path in it and may need migration
class ClassificationMistakennessConfig(MistakennessMethodConfig):
    def __init__(
        self, pred_field, label_field, mistakenness_field, use_logits, **kwargs
    ):
        super().__init__(pred_field, label_field, mistakenness_field, **kwargs)
        self.use_logits = use_logits

    @property
    def method(self):
        return "classification"


class ClassificationMistakenness(MistakennessMethod):
    def process_image(self, sample_or_frame):
        use_logits = self.config.use_logits

        pred_label, gt_label = _get_data(
            sample_or_frame, self.pred_field, self.label_field, use_logits
        )

        if pred_label is None and gt_label is None:
            return None

        if pred_label is None or gt_label is None:
            m = 1.0
        elif isinstance(pred_label, fol.Classifications):
            # For multilabel problems, all labels must match
            pred_labels = set(c.label for c in pred_label.classifications)
            gt_labels = set(c.label for c in gt_label.classifications)
            m = float(pred_labels == gt_labels)
        else:
            m = float(pred_label.label == gt_label.label)

        if pred_label is None:
            mistakenness = 1.0
        elif use_logits:
            mistakenness = _compute_mistakenness_class(pred_label.logits, m)
        else:
            mistakenness = _compute_mistakenness_class_conf(
                pred_label.confidence, m
            )

        return mistakenness

    def get_fields(self, samples, brain_key):
        pred_field = self.config.pred_field
        label_field = self.config.label_field
        mistakenness_field = self.config.mistakenness_field

        fields = [pred_field, label_field, mistakenness_field]

        if samples._is_frame_field(label_field):
            fields.append(samples._FRAMES_PREFIX + mistakenness_field)

        return fields

    def cleanup(self, samples, brain_key):
        label_field = self.config.label_field
        mistakenness_field = self.config.mistakenness_field

        samples._dataset.delete_sample_fields(
            mistakenness_field, error_level=1
        )

        if samples._is_frame_field(label_field):
            samples._dataset.delete_frame_fields(
                mistakenness_field, error_level=1
            )


# @todo move to `fiftyone/brain/mistakenness.py`
# Don't do this hastily; `get_brain_info()` on existing datasets has this
# class's full path in it and may need migration
class DetectionMistakennessConfig(MistakennessMethodConfig):
    def __init__(
        self,
        pred_field,
        label_field,
        mistakenness_field,
        missing_field,
        spurious_field,
        use_logits,
        copy_missing,
        eval_key,
        **kwargs
    ):
        super().__init__(pred_field, label_field, mistakenness_field, **kwargs)
        self.missing_field = missing_field
        self.spurious_field = spurious_field
        self.use_logits = use_logits
        self.copy_missing = copy_missing
        self.eval_key = eval_key

    @property
    def method(self):
        return "detection"


class DetectionMistakenness(MistakennessMethod):
    def process_image(self, sample_or_frame, eval_key):
        missing_field = self.config.missing_field
        spurious_field = self.config.spurious_field
        mistakenness_field = self.config.mistakenness_field
        copy_missing = self.config.copy_missing
        use_logits = self.config.use_logits

        pred_label, gt_label = _get_data(
            sample_or_frame, self.pred_field, self.label_field, use_logits
        )

        list_field = self.label_type._LABEL_LIST_FIELD

        if pred_label is None:
            pred_label = self.label_type()

        if gt_label is None:
            gt_label = self.label_type()

        num_spurious = 0
        num_missing = 0
        missing_objects = {}
        image_mistakenness = []
        pred_map = {}
        for pred_obj in pred_label[list_field]:
            pred_map[pred_obj.id] = pred_obj
            gt_id = pred_obj[eval_key + "_id"]
            conf = pred_obj.confidence
            if gt_id == "" and conf > _MISSED_CONFIDENCE_THRESHOLD:
                # Unmached FP with high confidence are missing
                pred_obj[missing_field] = True
                num_missing += 1
                missing_objects[pred_obj.id] = pred_obj

        for gt_obj in gt_label[list_field]:
            # Avoid adding the same unmatched FP predictions upon multiple runs
            # of this method
            if copy_missing and gt_obj.has_field(missing_field):
                if gt_obj.id in missing_objects:
                    del missing_objects[gt_obj.id]

                continue

            pred_id = gt_obj[eval_key + "_id"]
            if pred_id == "":
                # FN may be spurious
                gt_obj[spurious_field] = True
                num_spurious += 1
            else:
                # For matched FP, compute mistakenness
                iou = gt_obj[eval_key + "_iou"]
                pred_obj = pred_map[pred_id]
                m = float(gt_obj.label == pred_obj.label)
                if use_logits:
                    mistakenness_class = _compute_mistakenness_class(
                        pred_obj.logits, m
                    )
                    mistakenness_loc = _compute_mistakenness_loc(
                        pred_obj.logits, iou
                    )
                else:
                    mistakenness_class = _compute_mistakenness_class_conf(
                        pred_obj.confidence, m
                    )
                    mistakenness_loc = _compute_mistakenness_loc_conf(
                        pred_obj.confidence, iou
                    )

                gt_obj[mistakenness_field] = mistakenness_class
                gt_obj[mistakenness_field + "_loc"] = mistakenness_loc
                image_mistakenness.append(mistakenness_class)

        if copy_missing:
            gt_label[list_field].extend(missing_objects.values())
            sample_or_frame[self.label_field] = gt_label

        if image_mistakenness:
            mistakenness = np.max(image_mistakenness)
        else:
            mistakenness = -1

        return mistakenness, num_missing, num_spurious

    def get_fields(self, samples, brain_key):
        pred_field = self.config.pred_field
        label_field = self.config.label_field
        mistakenness_field = self.config.mistakenness_field
        missing_field = self.config.missing_field
        spurious_field = self.config.spurious_field

        label_type = samples._get_label_field_type(pred_field)
        list_field = label_type._LABEL_LIST_FIELD

        fields = [
            mistakenness_field,
            missing_field,
            spurious_field,
            "%s.%s.%s" % (label_field, list_field, mistakenness_field),
            "%s.%s.%s_loc" % (label_field, list_field, mistakenness_field),
            "%s.%s.%s" % (pred_field, list_field, missing_field),
            "%s.%s.%s" % (label_field, list_field, spurious_field),
        ]

        if samples._is_frame_field(pred_field):
            fields.extend(
                [
                    samples._FRAMES_PREFIX + mistakenness_field,
                    samples._FRAMES_PREFIX + missing_field,
                    samples._FRAMES_PREFIX + spurious_field,
                ]
            )

        return fields

    def cleanup(self, samples, brain_key):
        pred_field = self.config.pred_field
        label_field = self.config.label_field
        mistakenness_field = self.config.mistakenness_field
        missing_field = self.config.missing_field
        spurious_field = self.config.spurious_field
        eval_key = self.config.eval_key

        label_type = samples._get_label_field_type(pred_field)
        list_field = label_type._LABEL_LIST_FIELD

        pred_field, is_frame_field = samples._handle_frame_field(pred_field)
        label_field, _ = samples._handle_frame_field(label_field)

        fields = [
            mistakenness_field,
            missing_field,
            spurious_field,
            "%s.%s.%s" % (label_field, list_field, mistakenness_field),
            "%s.%s.%s_loc" % (label_field, list_field, mistakenness_field),
            "%s.%s.%s" % (pred_field, list_field, missing_field),
            "%s.%s.%s" % (label_field, list_field, spurious_field),
        ]

        if self.config.copy_missing:
            # Remove objects that were added to `label_field`
            samples._dataset.filter_labels(
                self.config.label_field, F(missing_field).exists(False)
            ).save()

        if is_frame_field:
            samples._dataset.delete_sample_fields(
                [mistakenness_field, spurious_field, missing_field],
                error_level=1,
            )
            samples._dataset.delete_frame_fields(fields, error_level=1)
        else:
            samples._dataset.delete_sample_fields(fields, error_level=1)

        if eval_key in samples.list_evaluations():
            samples.delete_evaluation(eval_key)

    def _validate_run(self, samples, brain_key, existing_info):
        super()._validate_run(samples, brain_key, existing_info)
        self._validate_fields_match(brain_key, "missing_field", existing_info)
        self._validate_fields_match(brain_key, "spurious_field", existing_info)
        self._validate_fields_match(brain_key, "copy_missing", existing_info)


def _make_eval_key(samples, brain_key):
    existing_eval_keys = samples.list_evaluations()
    eval_key = brain_key + "_eval"
    if eval_key not in existing_eval_keys:
        return eval_key

    idx = 2
    while eval_key + str(idx) in existing_eval_keys:
        idx += 1

    return eval_key + str(idx)


def _get_data(sample, pred_field, label_field, use_logits):
    pred_label = sample[pred_field]
    label = sample[label_field]

    if pred_label is None:
        return pred_label, label

    if isinstance(pred_label, fol.Detections):
        for det in pred_label.detections:
            if det.confidence is None:
                raise ValueError(
                    "Detection '%s' in sample '%s' field '%s' has no "
                    "confidence" % (det.id, sample.id, pred_field)
                )
    elif isinstance(pred_label, fol.Polylines):
        for poly in pred_label.polylines:
            if poly.confidence is None:
                raise ValueError(
                    "Polyline '%s' in sample '%s' field '%s' has no "
                    "confidence" % (poly.id, sample.id, pred_field)
                )
    elif use_logits:
        if pred_label.logits is None:
            raise ValueError(
                "Sample '%s' field '%s' has no logits"
                % (sample.id, pred_field)
            )
    else:
        if pred_label.confidence is None:
            raise ValueError(
                "Sample '%s' field '%s' has no confidence"
                % (sample.id, pred_field)
            )

    return pred_label, label


def _compute_mistakenness_class(logits, m):
    # constrain m to either 1 (incorrect) or -1 (correct)
    m = m * -2.0 + 1.0

    c = -1.0 * entropy(softmax(np.asarray(logits)))
    mistakenness = (m * exp(c) + 1.0) / 2.0

    return mistakenness


def _compute_mistakenness_loc(logits, iou):
    # i = 0 for high iou, i = 1 for low iou
    i = (1.0 / (1.0 - _DETECTION_IOU)) * (1.0 - iou)

    # c = 0 for low confidence, c = 1 for high confidence
    c = exp(-1.0 * entropy(softmax(np.asarray(logits))))

    # mistakenness = i when c = i, mistakenness = 0.5 if c = 0
    # mistakenness is higher with lower IoU and closer to 0 or 1 with higher
    # confidence
    mistakenness = (c * ((2.0 * i) - 1.0) + 1.0) / 2.0

    return mistakenness


def _compute_mistakenness_class_conf(confidence, m):
    # constrain m to either 1 (incorrect) or -1 (correct)
    m = m * -2.0 + 1.0

    mistakenness = (m * confidence + 1.0) / 2.0

    return mistakenness


def _compute_mistakenness_loc_conf(confidence, iou):
    # i = 0 for high iou, i = 1 for low iou
    i = (1.0 / (1.0 - _DETECTION_IOU)) * (1.0 - iou)

    # c = 0 for low confidence, c = 1 for high confidence
    c = confidence

    # mistakenness = i when c = i, mistakenness = 0.5 if c = 0
    # mistakenness is higher with lower IoU and closer to 0 or 1 with higher
    # confidence
    mistakenness = (c * ((2.0 * i) - 1.0) + 1.0) / 2.0

    return mistakenness



"""
Visualization interface.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
from copy import deepcopy
import inspect
import logging

import numpy as np
import sklearn.decomposition as skd
import sklearn.manifold as skm

import eta.core.utils as etau

import fiftyone.brain as fb
import fiftyone.core.brain as fob
import fiftyone.core.expressions as foe
import fiftyone.core.fields as fof
import fiftyone.core.plots as fop
import fiftyone.core.utils as fou
import fiftyone.core.validation as fov

fbu = fou.lazy_import("fiftyone.brain.internal.core.utils")

umap = fou.lazy_import("umap")


logger = logging.getLogger(__name__)

_DEFAULT_MODEL = "mobilenet-v2-imagenet-torch"
_DEFAULT_BATCH_SIZE = None


def compute_visualization(
    samples,
    patches_field,
    embeddings,
    points,
    create_index,
    points_field,
    brain_key,
    num_dims,
    method,
    similarity_index,
    model,
    model_kwargs,
    force_square,
    alpha,
    batch_size,
    num_workers,
    skip_failures,
    progress,
    **kwargs,
):
    """See ``fiftyone/brain/__init__.py``."""

    fov.validate_collection(samples)

    if method == "manual" and points is None:
        raise ValueError(
            "You must provide your own `points` when `method='manual'`"
        )

    if points is not None:
        method = "manual"
        model = None
        embeddings = None
        embeddings_field = None
        num_dims = _get_dimension(points)

    if create_index and points_field is None:
        points_field = brain_key

    if points_field is not None and num_dims != 2:
        raise ValueError("`points_field` is only supported when `num_dims=2`")

    if etau.is_str(embeddings):
        embeddings_field, embeddings_exist = fbu.parse_data_field(
            samples,
            embeddings,
            patches_field=patches_field,
            data_type="embeddings",
        )
        embeddings = None
    else:
        embeddings_field = None
        embeddings_exist = None

    if points_field is not None:
        points_field, _ = fbu.parse_data_field(
            samples,
            points_field,
            patches_field=patches_field,
            data_type="points",
        )

    if etau.is_str(similarity_index):
        similarity_index = samples.load_brain_results(similarity_index)

    if (
        model is None
        and points is None
        and embeddings is None
        and similarity_index is None
        and not embeddings_exist
    ):
        model = _DEFAULT_MODEL
        if batch_size is None:
            batch_size = _DEFAULT_BATCH_SIZE

    config = _parse_config(
        method,
        embeddings_field=embeddings_field,
        points_field=points_field,
        similarity_index=similarity_index,
        model=model,
        model_kwargs=model_kwargs,
        patches_field=patches_field,
        num_dims=num_dims,
        **kwargs,
    )

    brain_method = config.build()
    brain_method.ensure_requirements()

    if brain_key is not None:
        brain_method.register_run(samples, brain_key)

    if points is None:
        embeddings, sample_ids, label_ids = fbu.get_embeddings(
            samples,
            model=model,
            model_kwargs=model_kwargs,
            patches_field=patches_field,
            embeddings_field=embeddings_field,
            embeddings=embeddings,
            similarity_index=similarity_index,
            force_square=force_square,
            alpha=alpha,
            batch_size=batch_size,
            num_workers=num_workers,
            skip_failures=skip_failures,
            progress=progress,
        )

        logger.info("Generating visualization...")
        points = brain_method.fit(embeddings)
    else:
        points, sample_ids, label_ids = fbu.parse_data(
            samples,
            patches_field=patches_field,
            data=points,
            data_type="points",
        )

    if points_field is not None:
        _generate_spatial_index(
            samples,
            points,
            points_field,
            sample_ids,
            label_ids=label_ids,
            patches_field=patches_field,
            create_index=create_index,
            progress=progress,
        )

    results = VisualizationResults(
        samples,
        config,
        brain_key,
        points,
        sample_ids=sample_ids,
        label_ids=label_ids,
    )

    brain_method.save_run_results(samples, brain_key, results)

    return results


def values(results, path_or_expr):
    samples = results.view
    patches_field = results.config.patches_field
    if patches_field is not None:
        ids = results.current_label_ids
    else:
        ids = results.current_sample_ids

    return fbu.get_values(
        samples, path_or_expr, ids, patches_field=patches_field
    )


def visualize(
    results,
    labels=None,
    sizes=None,
    classes=None,
    backend="plotly",
    **kwargs,
):
    points = results.current_points
    samples = results.view
    patches_field = results.config.patches_field
    good_inds = results._curr_good_inds
    if patches_field is not None:
        ids = results.current_label_ids
    else:
        ids = results.current_sample_ids

    if good_inds is not None:
        if etau.is_container(labels) and not _is_expr(labels):
            labels = fbu.filter_values(
                labels, good_inds, patches_field=patches_field
            )

        if etau.is_container(sizes) and not _is_expr(sizes):
            sizes = fbu.filter_values(
                sizes, good_inds, patches_field=patches_field
            )

    if labels is not None and _is_expr(labels):
        labels = fbu.get_values(
            samples, labels, ids, patches_field=patches_field
        )

    if sizes is not None and _is_expr(sizes):
        sizes = fbu.get_values(
            samples, sizes, ids, patches_field=patches_field
        )

    return fop.scatterplot(
        points,
        samples=samples,
        ids=ids,
        link_field=patches_field,
        labels=labels,
        sizes=sizes,
        classes=classes,
        backend=backend,
        **kwargs,
    )


def _is_expr(arg):
    return isinstance(arg, (foe.ViewExpression, dict))


def _parse_config(name, **kwargs):
    if name is None:
        name = fb.brain_config.default_visualization_method

    if inspect.isclass(name):
        return name(**kwargs)

    methods = fb.brain_config.visualization_methods

    if name not in methods:
        raise ValueError(
            "Unsupported method '%s'. The available methods are %s"
            % (name, sorted(methods.keys()))
        )

    params = deepcopy(methods[name])

    config_cls = kwargs.pop("config_cls", None)

    if config_cls is None:
        config_cls = params.pop("config_cls", None)

    if config_cls is None:
        raise ValueError(
            "Visualization method '%s' has no `config_cls`" % name
        )

    if etau.is_str(config_cls):
        config_cls = etau.get_class(config_cls)

    params.update(**kwargs)
    return config_cls(**params)


def _get_dimension(points):
    if isinstance(points, dict):
        points = next(iter(points.values()), None)

    if isinstance(points, list):
        points = next(iter(points), None)

    if points is None:
        return 2

    return points.shape[-1]


def _generate_spatial_index(
    samples,
    points,
    points_field,
    sample_ids,
    label_ids=None,
    patches_field=None,
    create_index=True,
    progress=False,
):
    # Indexes are not currently usable on patch visualizations
    if create_index and patches_field is not None:
        create_index = False

    dataset = samples._root_dataset
    if patches_field is not None:
        _, points_field = dataset._get_label_field_path(
            patches_field, points_field
        )

    logger.info("Generating spatial index in field '%s'...", points_field)

    dataset.add_sample_field(
        points_field, fof.ListField, subfield=fof.FloatField
    )

    points = points.astype(float)

    if create_index:
        min_val, max_val = points.min(), points.max()
        dataset.create_index([(points_field, "2d")], min=min_val, max=max_val)

    points = points.tolist()
    if patches_field is not None:
        values = dict(zip(label_ids, points))
        dataset.set_label_values(points_field, values, progress=progress)
    else:
        values = dict(zip(sample_ids, points))
        dataset.set_values(
            points_field, values, key_field="id", progress=progress
        )


class VisualizationResults(fob.BrainResults):
    """Class storing the results of
    :meth:`fiftyone.brain.compute_visualization`.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`VisualizationConfig` used
        brain_key: the brain key
        points: a ``num_points x num_dims`` array of visualization points
        sample_ids (None): a ``num_points`` array of sample IDs
        label_ids (None): a ``num_points`` array of label IDs, if applicable
        backend (None): a :class:`Visualization` backend
    """

    def __init__(
        self,
        samples,
        config,
        brain_key,
        points,
        sample_ids=None,
        label_ids=None,
        backend=None,
    ):
        super().__init__(samples, config, brain_key, backend=backend)

        if sample_ids is None:
            sample_ids, label_ids = fbu.get_ids(
                samples,
                patches_field=config.patches_field,
                data=points,
                data_type="points",
            )

        self.points = points
        self.sample_ids = sample_ids
        self.label_ids = label_ids

        self._last_view = None
        self._curr_view = None
        self._curr_points = None
        self._curr_sample_ids = None
        self._curr_label_ids = None
        self._curr_keep_inds = None
        self._curr_good_inds = None

        self.use_view(samples)

    def __enter__(self):
        self._last_view = self.view
        return self

    def __exit__(self, *args):
        self.use_view(self._last_view)
        self._last_view = None

    @property
    def config(self):
        """The :class:`VisualizationConfig` for the results."""
        return self._config

    @property
    def index_size(self):
        """The number of active points in the index.

        If :meth:`use_view` has been called to restrict the index, this
        property will reflect the size of the active index.
        """
        return len(self._curr_sample_ids)

    @property
    def total_index_size(self):
        """The total number of data points in the index.

        If :meth:`use_view` has been called to restrict the index, this value
        may be larger than the current :meth:`index_size`.
        """
        return len(self.points)

    @property
    def missing_size(self):
        """The total number of data points in :meth:`view` that are missing
        from this index.

        This property is only applicable when :meth:`use_view` has been called,
        and it will be ``None`` if no data points are missing.
        """
        good = self._curr_good_inds

        if good is None:
            return None

        return good.size - np.count_nonzero(good)

    @property
    def current_points(self):
        """The currently active points in the index.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_points

    @property
    def current_sample_ids(self):
        """The sample IDs of the currently active points in the index.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_sample_ids

    @property
    def current_label_ids(self):
        """The label IDs of the currently active points in the index, or
        ``None`` if not applicable.

        If :meth:`use_view` has been called, this may be a subset of the full
        index.
        """
        return self._curr_label_ids

    @property
    def view(self):
        """The :class:`fiftyone.core.collections.SampleCollection` against
        which results are currently being generated.

        If :meth:`use_view` has been called, this view may be different than
        the collection on which the full index was generated.
        """
        return self._curr_view

    @property
    def has_spatial_index(self):
        """Whether these results have a spatial index.

        Use :meth:`index_points` to add a spatial index to an existing set of
        visualization results.
        """
        return self.config.points_field is not None

    def use_view(
        self, sample_collection, allow_missing=True, warn_missing=False
    ):
        """Restricts the index to the provided view.

        Subsequent calls to methods on this instance will only contain results
        from the specified view rather than the full index.

        Use :meth:`clear_view` to reset to the full index. Or, equivalently,
        use the context manager interface as demonstrated below to
        automatically reset the view when the context exits.

        Example usage::

            import fiftyone as fo
            import fiftyone.brain as fob
            import fiftyone.zoo as foz

            dataset = foz.load_zoo_dataset("quickstart")

            results = fob.compute_visualization(dataset)
            print(results.index_size)  # 200

            view = dataset.take(50)

            with results.use_view(view):
                print(results.index_size)  # 50

                plot = results.visualize()
                plot.show()

        Args:
            sample_collection: a
                :class:`fiftyone.core.collections.SampleCollection`
            allow_missing (True): whether to allow the provided collection to
                contain data points that this index does not contain (True) or
                whether to raise an error in this case (False)
            warn_missing (False): whether to log a warning if the provided
                collection contains data points that this index does not
                contain

        Returns:
            self
        """
        sample_ids, label_ids, keep_inds, good_inds = fbu.filter_ids(
            sample_collection,
            self.sample_ids,
            self.label_ids,
            patches_field=self._config.patches_field,
            allow_missing=allow_missing,
            warn_missing=warn_missing,
        )

        if keep_inds is not None:
            points = self.points[keep_inds, :]
        else:
            points = self.points

        self._curr_view = sample_collection
        self._curr_points = points
        self._curr_sample_ids = sample_ids
        self._curr_label_ids = label_ids
        self._curr_keep_inds = keep_inds
        self._curr_good_inds = good_inds

        return self

    def clear_view(self):
        """Clears the view set by :meth:`use_view`, if any.

        Subsequent operations will be performed on the full index.
        """
        self.use_view(self._samples)

    def values(self, path_or_expr):
        """Extracts a flat list of values from the given field or expression
        corresponding to the current :meth:`view`.

        This method always returns values in the same order as
        :meth:`current_points`, :meth:`current_sample_ids`, and
        :meth:`current_label_ids`.

        Args:
            path_or_expr: the values to extract, which can be:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric or string values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric or string values to compute via
                    :meth:`fiftyone.core.collections.SampleCollection.values`

        Returns:
            a list of values
        """
        return values(self, path_or_expr)

    def visualize(
        self,
        labels=None,
        sizes=None,
        classes=None,
        backend="plotly",
        **kwargs,
    ):
        """Generates an interactive scatterplot of the visualization results
        for the current :meth:`view`.

        This method supports 2D or 3D visualizations, but interactive point
        selection is only available in 2D.

        You can use the ``labels`` parameters to define a coloring for the
        points, and you can use the ``sizes`` parameter to scale the sizes of
        the points.

        You can attach plots generated by this method to an App session via its
        :attr:`fiftyone.core.session.Session.plots` attribute, which will
        automatically sync the session's view with the currently selected
        points in the plot.

        Args:
            labels (None): data to use to color the points. Can be any of the
                following:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric or string values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric or string values to compute via
                    :meth:`fiftyone.core.collections.SampleCollection.values`
                -   a list or array-like of numeric or string values
                -   a list of lists of numeric or string values, if the data in
                    this visualization corresponds to a label list field like
                    :class:`fiftyone.core.labels.Detections`

            sizes (None): data to use to scale the sizes of the points. Can be
                any of the following:

                -   the name of a sample field or ``embedded.field.name`` from
                    which to extract numeric values
                -   a :class:`fiftyone.core.expressions.ViewExpression`
                    defining numeric values to compute via
                    :meth:`fiftyone.core.collections.SampleCollection.values`
                -   a list or array-like of numeric values
                -   a list of lists of numeric values, if the data in this
                    visualization corresponds to a label list field like
                    :class:`fiftyone.core.labels.Detections`

            classes (None): an optional list of classes whose points to plot.
                Only applicable when ``labels`` contains strings
            backend ("plotly"): the plotting backend to use. Supported values
                are ``("plotly", "matplotlib")``
            **kwargs: keyword arguments for the backend plotting method:

                -   "plotly" backend: :meth:`fiftyone.core.plots.plotly.scatterplot`
                -   "matplotlib" backend: :meth:`fiftyone.core.plots.matplotlib.scatterplot`

        Returns:
            an :class:`fiftyone.core.plots.base.InteractivePlot`
        """
        return visualize(
            self,
            labels=labels,
            sizes=sizes,
            classes=classes,
            backend=backend,
            **kwargs,
        )

    def index_points(
        self,
        points_field=None,
        create_index=True,
        progress=None,
    ):
        """Adds a spatial index for these visualization results to its
        dataset's samples.

        This method is useful if you want to add a spatial index to existing
        visualization results that don't yet have one.

        Spatial indexes are highly recommended for large datasets as they
        enable efficient querying when lassoing points in embeddings plots.

        Args:
            points_field (None): an optional field name in which to store the
                spatial index. The default is the result's ``brain_key``
            create_index (True): whether to create a database index for the
                points
            progress (None): whether to render a progress bar (True/False),
                use the default value ``fiftyone.config.show_progress_bars``
                (None), or a progress callback function to invoke instead
        """
        if points_field is None:
            if self.key is None:
                raise ValueError(
                    "You must provide a `points_field` when indexing points "
                    "that are not associated with a brain key"
                )

            points_field = self.key

        _generate_spatial_index(
            self.samples,
            self.points,
            points_field,
            self.sample_ids,
            label_ids=self.label_ids,
            patches_field=self.config.patches_field,
            create_index=create_index,
            progress=progress,
        )

        if self.key is not None:
            self.config.points_field = points_field
            self.save_config()

    def remove_index(self):
        """Removes the spatial index from these visualization results, if one
        exists.
        """
        points_field = self.config.points_field
        if points_field is None:
            return

        dataset = self.samples._root_dataset
        if self.config.patches_field is not None:
            _, points_field = dataset._get_label_field_path(
                self.config.patches_field, points_field
            )

        dataset.delete_sample_field(points_field, error_level=1)

        if self.key is not None:
            self.config.points_field = None
            self.save_config()

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        points = np.array(d["points"])

        sample_ids = d.get("sample_ids", None)
        if sample_ids is not None:
            sample_ids = np.array(sample_ids)

        label_ids = d.get("label_ids", None)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return cls(
            samples,
            config,
            brain_key,
            points,
            sample_ids=sample_ids,
            label_ids=label_ids,
        )


class VisualizationConfig(fob.BrainMethodConfig):
    """Base class for configuring visualization methods.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        points_field (None): the name of a field in which to store the
            visualization points, if requested
        similarity_index (None): the similarity index containing the
            embeddings, if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        model_kwargs (None): a dictionary of optional keyword arguments to pass
            to the model's ``Config`` when a model name is provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        num_dims (2): the dimension of the visualization space
    """

    def __init__(
        self,
        embeddings_field=None,
        points_field=None,
        similarity_index=None,
        model=None,
        model_kwargs=None,
        patches_field=None,
        num_dims=2,
        **kwargs,
    ):
        if similarity_index is not None and not etau.is_str(similarity_index):
            similarity_index = similarity_index.key

        if model is not None and not etau.is_str(model):
            model = None

        self.embeddings_field = embeddings_field
        self.points_field = points_field
        self.similarity_index = similarity_index
        self.model = model
        self.model_kwargs = model_kwargs
        self.patches_field = patches_field
        self.num_dims = num_dims
        super().__init__(**kwargs)

    @property
    def type(self):
        return "visualization"


class Visualization(fob.BrainMethod):
    def fit(self, embeddings):
        raise NotImplementedError("subclass must implement fit()")

    def get_fields(self, samples, brain_key):
        fields = []
        if self.config.patches_field is not None:
            fields.append(self.config.patches_field)
        elif self.config.points_field is not None:
            fields.append(self.config.points_field)

        return fields

    def rename(self, samples, key, new_key):
        patches_field = self.config.patches_field
        points_field = self.config.points_field
        dataset = samples._root_dataset

        if points_field is not None and points_field == key:
            old_path = key
            new_path = new_key
            if patches_field is not None:
                _, old_path = dataset._get_label_field_path(
                    patches_field, old_path
                )
                _, new_path = dataset._get_label_field_path(
                    patches_field, new_path
                )

            self.config.points_field = new_key
            self.update_run_config(samples, key, self.config)

            dataset.rename_sample_field(old_path, new_path)

    def cleanup(self, samples, key):
        patches_field = self.config.patches_field
        points_field = self.config.points_field
        dataset = samples._root_dataset

        if points_field is not None:
            if patches_field is not None:
                _, points_field = dataset._get_label_field_path(
                    patches_field, points_field
                )

            dataset.delete_sample_field(points_field, error_level=1)


class UMAPVisualizationConfig(VisualizationConfig):
    """Configuration for Uniform Manifold Approximation and Projection (UMAP)
    embedding visualization.

    See https://github.com/lmcinnes/umap for more information about the
    supported parameters.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        points_field (None): the name of a field in which to store the
            visualization points, if requested
        similarity_index (None): the similarity index containing the
            embeddings, if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        model_kwargs (None): a dictionary of optional keyword arguments to pass
            to the model's ``Config`` when a model name is provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        num_dims (2): the dimension of the visualization space
        num_neighbors (15): the number of neighboring points used in local
            approximations of manifold structure. Larger values will result in
            more global structure being preserved at the loss of detailed local
            structure. Typical values are in ``[5, 50]``
        metric ("euclidean"): the metric to use when calculating distance
            between embeddings. See the UMAP documentation for supported values
        min_dist (0.1): the effective minimum distance between embedded
            points. This controls how tightly the embedding is allowed compress
            points together. Larger values ensure embedded points are more
            evenly distributed, while smaller values allow the algorithm to
            optimise more accurately with regard to local structure. Typical
            values are in ``[0.001, 0.5]``
        seed (None): a random seed
        verbose (True): whether to log progress
    """

    def __init__(
        self,
        embeddings_field=None,
        points_field=None,
        similarity_index=None,
        model=None,
        model_kwargs=None,
        patches_field=None,
        num_dims=2,
        num_neighbors=15,
        metric="euclidean",
        min_dist=0.1,
        seed=None,
        verbose=True,
        **kwargs,
    ):
        super().__init__(
            embeddings_field=embeddings_field,
            points_field=points_field,
            similarity_index=similarity_index,
            model=model,
            model_kwargs=model_kwargs,
            patches_field=patches_field,
            num_dims=num_dims,
            **kwargs,
        )
        self.num_neighbors = num_neighbors
        self.metric = metric
        self.min_dist = min_dist
        self.seed = seed
        self.verbose = verbose

    @property
    def method(self):
        return "umap"


class UMAPVisualization(Visualization):
    def ensure_requirements(self):
        fou.ensure_package(
            "umap-learn>=0.5",
            error_msg=(
                "You must install the `umap-learn>=0.5` package in order to "
                "use UMAP-based visualization. This is recommended, as UMAP "
                "is awesome! If you do not wish to install UMAP, try "
                "`method='tsne'` instead"
            ),
        )

    def fit(self, embeddings):
        _umap = umap.UMAP(
            n_components=self.config.num_dims,
            n_neighbors=self.config.num_neighbors,
            metric=self.config.metric,
            min_dist=self.config.min_dist,
            random_state=self.config.seed,
            verbose=self.config.verbose,
        )
        return _umap.fit_transform(embeddings)


class TSNEVisualizationConfig(VisualizationConfig):
    """Configuration for t-distributed Stochastic Neighbor Embedding (t-SNE)
    visualization.

    See https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html
    for more information about the supported parameters.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        points_field (None): the name of a field in which to store the
            visualization points, if requested
        similarity_index (None): the similarity index containing the
            embeddings, if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        model_kwargs (None): a dictionary of optional keyword arguments to pass
            to the model's ``Config`` when a model name is provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        num_dims (2): the dimension of the visualization space
        pca_dims (50): the number of PCA dimensions to compute prior to running
            t-SNE. It is highly recommended to reduce the number of dimensions
            to a reasonable number (e.g. 50) before running t-SNE, as this will
            suppress some noise and speed up the computation of pairwise
            distances between samples
        svd_solver ("randomized"): the SVD solver to use when performing PCA.
            Consult the sklearn docmentation for details
        metric ("euclidean"): the metric to use when calculating distance
            between embeddings. Must be a supported value for the ``metric``
            argument of ``scipy.spatial.distance.pdist``
        perplexity (30.0): the perplexity to use. Perplexity is related to the
            number of nearest neighbors that is used in other manifold learning
            algorithms. Larger datasets usually require a larger perplexity.
            Typical values are in ``[5, 50]``
        learning_rate (200.0): the learning rate to use. Typical values are
            in ``[10, 1000]``. If the learning rate is too high, the data may
            look like a ball with any point approximately equidistant from its
            nearest neighbours. If the learning rate is too low, most points
            may look compressed in a dense cloud with few outliers. If the cost
            function gets stuck in a bad local minimum increasing the learning
            rate may help
        max_iters (1000): the maximum number of iterations to run. Should be at
            least 250
        seed (None): a random seed
        verbose (True): whether to log progress
    """

    def __init__(
        self,
        embeddings_field=None,
        points_field=None,
        similarity_index=None,
        model=None,
        model_kwargs=None,
        patches_field=None,
        num_dims=2,
        pca_dims=50,
        svd_solver="randomized",
        metric="euclidean",
        perplexity=30.0,
        learning_rate=200.0,
        max_iters=1000,
        seed=None,
        verbose=True,
        **kwargs,
    ):
        super().__init__(
            embeddings_field=embeddings_field,
            points_field=points_field,
            similarity_index=similarity_index,
            model=model,
            model_kwargs=model_kwargs,
            patches_field=patches_field,
            num_dims=num_dims,
            **kwargs,
        )
        self.pca_dims = pca_dims
        self.svd_solver = svd_solver
        self.metric = metric
        self.perplexity = perplexity
        self.learning_rate = learning_rate
        self.max_iters = max_iters
        self.seed = seed
        self.verbose = verbose

    @property
    def method(self):
        return "tsne"


class TSNEVisualization(Visualization):
    def fit(self, embeddings):
        if self.config.pca_dims is not None:
            _pca = skd.PCA(
                n_components=self.config.pca_dims,
                svd_solver=self.config.svd_solver,
                random_state=self.config.seed,
            )
            embeddings = _pca.fit_transform(embeddings)

        embeddings = embeddings.astype(np.float32, copy=False)

        verbose = 2 if self.config.verbose else 0

        _tsne = skm.TSNE(
            n_components=self.config.num_dims,
            perplexity=self.config.perplexity,
            learning_rate=self.config.learning_rate,
            metric=self.config.metric,
            init="pca",  # "random" or "pca"
            n_iter=self.config.max_iters,
            random_state=self.config.seed,
            verbose=verbose,
        )
        return _tsne.fit_transform(embeddings)


class PCAVisualizationConfig(VisualizationConfig):
    """Configuration for principal component analysis (PCA) embedding
    visualization.

    See https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html
    for more information about the supported parameters.

    Args:
        embeddings_field (None): the sample field containing the embeddings,
            if one was provided
        points_field (None): the name of a field in which to store the
            visualization points, if requested
        similarity_index (None): the similarity index containing the
            embeddings, if one was provided
        model (None): the :class:`fiftyone.core.models.Model` or name of the
            zoo model that was used to compute embeddings, if known
        model_kwargs (None): a dictionary of optional keyword arguments to pass
            to the model's ``Config`` when a model name is provided
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        num_dims (2): the dimension of the visualization space
        svd_solver ("randomized"): the SVD solver to use. Consult the sklearn
            docmentation for details
        seed (None): a random seed
    """

    def __init__(
        self,
        embeddings_field=None,
        points_field=None,
        similarity_index=None,
        model=None,
        model_kwargs=None,
        patches_field=None,
        num_dims=2,
        svd_solver="randomized",
        seed=None,
        **kwargs,
    ):
        super().__init__(
            embeddings_field=embeddings_field,
            points_field=points_field,
            similarity_index=similarity_index,
            model=model,
            model_kwargs=model_kwargs,
            patches_field=patches_field,
            num_dims=num_dims,
            **kwargs,
        )
        self.svd_solver = svd_solver
        self.seed = seed

    @property
    def method(self):
        return "pca"


class PCAVisualization(Visualization):
    def fit(self, embeddings):
        _pca = skd.PCA(
            n_components=self.config.num_dims,
            svd_solver=self.config.svd_solver,
            random_state=self.config.seed,
        )
        return _pca.fit_transform(embeddings)


class ManualVisualizationConfig(VisualizationConfig):
    """Configuration for manually-provided low-dimensional visualizations.

    Args:
        patches_field (None): the sample field defining the patches being
            analyzed, if any
        num_dims (2): the dimension of the visualization space
    """

    def __init__(self, patches_field=None, num_dims=2, **kwargs):
        super().__init__(
            patches_field=patches_field, num_dims=num_dims, **kwargs
        )

    @property
    def method(self):
        return "manual"


class ManualVisualization(Visualization):
    def fit(self, embeddings):
        raise NotImplementedError(
            "The low-dimensional representation must be manually provided "
            "when using this method"
        )



"""
Uniqueness tests.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import unittest

import fiftyone as fo
import fiftyone.brain as fob
import fiftyone.zoo as foz


def test_uniqueness():
    _run_uniqueness()


def test_uniqueness_torch():
    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    _run_uniqueness(model=model, batch_size=16)


def test_uniqueness_tf():
    model = foz.load_zoo_model("resnet-v2-50-imagenet-tf1")
    _run_uniqueness(model=model, batch_size=16)


def test_uniqueness_missing():
    dataset = fo.Dataset()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    fob.compute_uniqueness(dataset, batch_size=1)

    view = dataset.exists("uniqueness")

    assert dataset.has_field("uniqueness")
    assert len(view) == 0


def test_roi_uniqueness():
    _run_uniqueness(roi_field="ground_truth")


def test_roi_uniqueness_torch():
    model = foz.load_zoo_model("inception-v3-imagenet-torch")
    _run_uniqueness(roi_field="ground_truth", model=model, batch_size=16)


def test_roi_uniqueness_tf():
    model = foz.load_zoo_model("resnet-v2-50-imagenet-tf1")
    _run_uniqueness(roi_field="ground_truth", model=model, batch_size=16)


def test_roi_uniqueness_missing():
    dataset = fo.Dataset()
    dataset.add_samples(
        [
            fo.Sample(filepath="non-existent1.png"),
            fo.Sample(filepath="non-existent2.png"),
            fo.Sample(filepath="non-existent3.png"),
            fo.Sample(filepath="non-existent4.png"),
        ]
    )

    for sample in dataset:
        sample["ground_truth"] = fo.Detections(
            detections=[fo.Detection(bounding_box=[0.1, 0.1, 0.8, 0.8])]
        )
        sample.save()

    fob.compute_uniqueness(dataset, roi_field="ground_truth", batch_size=1)

    view = dataset.exists("uniqueness")

    assert dataset.has_field("uniqueness")
    assert len(view) == 0


def test_uniqueness_similarity_index():
    dataset = foz.load_zoo_dataset(
        "quickstart", dataset_name=fo.get_default_dataset_name()
    )
    dataset.delete_sample_field("uniqueness")

    # Full similarity index

    similarity_index = fob.compute_similarity(
        dataset, brain_key="sklearn_index", backend="sklearn"
    )

    fob.compute_uniqueness(dataset, similarity_index=similarity_index)

    assert dataset.has_field("uniqueness")

    dataset.clear_cache()
    dataset.delete_sample_field("uniqueness")

    fob.compute_uniqueness(dataset, similarity_index="sklearn_index")

    assert dataset.has_field("uniqueness")

    # Partial similarity index

    view = dataset.take(100, seed=51)
    similarity_index2 = fob.compute_similarity(
        view, brain_key="sklearn_index2", backend="sklearn"
    )

    fob.compute_uniqueness(
        dataset,
        uniqueness_field="uniqueness2",
        similarity_index="sklearn_index2",
    )

    assert len(dataset.exists("uniqueness2")) == len(view)


def _run_uniqueness(roi_field=None, model=None, batch_size=None):
    dataset = foz.load_zoo_dataset(
        "quickstart", dataset_name=fo.get_default_dataset_name()
    )
    dataset.delete_sample_field("uniqueness")

    view = dataset.take(50)
    num_samples = len(view)

    fob.compute_uniqueness(
        view, roi_field=roi_field, model=model, batch_size=batch_size
    )

    num_uniqueness = dataset.count("uniqueness")
    assert num_uniqueness == num_samples

    bounds = dataset.bounds("uniqueness")

    assert bounds[0] >= 0
    assert bounds[1] <= 1


if __name__ == "__main__":
    fo.config.show_progress_bars = True
    unittest.main(verbosity=2)



"""
Mosaic similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

vector_search_client = fou.lazy_import("databricks.vector_search.client")


logger = logging.getLogger(__name__)

# Todo: add in required for arguments that are necessary to create the index table
class MosaicSimilarityConfig(SimilarityConfig):
    """Configuration for the Mosaic similarity backend.

    Args:
        endpoint_name (None): the name of the vector search endpoint that was created in the Databricks workspace
        workspace_url (None): the URL of the Databricks workspace
        catalog_name (None): the name of the catalog in the Databricks workspace
        schema_name (None): the name of the schema in the Databricks workspace
        index_name (None): the name of the index to use, if one is not provided, a unique name will be generated
        service_principal_client_id (None): the client ID of the service principal created for authentication
        service_principal_client_secret (None): the client secret of the service principal created for authentication
        personal_access_token (None): the personal access token created for authentication
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        endpoint_name=None,
        workspace_url=None,
        catalog_name=None,
        schema_name=None,
        index_name=None,
        service_principal_client_id=None,
        service_principal_client_secret=None,
        personal_access_token=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.index_name = index_name
        self.endpoint_name = endpoint_name
        self.catalog_name = catalog_name
        self.schema_name = schema_name

        # store privately so these aren't serialized
        self._workspace_url = workspace_url
        self._service_principal_client_id = service_principal_client_id
        self._service_principal_client_secret = service_principal_client_secret
        self._personal_access_token = personal_access_token

    @property
    def method(self):
        return "mosaic"

    @property
    def workspace_url(self):
        return self._workspace_url

    @workspace_url.setter
    def workspace_url(self, workspace_url):
        self._workspace_url = workspace_url

    @property
    def service_principal_client_id(self):
        return self._service_principal_client_id

    @service_principal_client_id.setter
    def service_principal_client_id(self, service_principal_client_id):
        self._service_principal_client_id = service_principal_client_id

    @property
    def service_principal_client_secret(self):
        return self._service_principal_client_secret

    @service_principal_client_secret.setter
    def service_principal_client_secret(self, service_principal_client_secret):
        self._service_principal_client_secret = service_principal_client_secret

    @property
    def personal_access_token(self):
        return self._personal_access_token

    @personal_access_token.setter
    def personal_access_token(self, personal_access_token):
        self._personal_access_token = personal_access_token

    @property
    def max_k(self):
        return None

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self,
        workspace_url=None,
        service_principal_client_id=None,
        service_principal_client_secret=None,
        personal_access_token=None,
    ):
        self._load_parameters(
            workspace_url=workspace_url,
            service_principal_client_id=service_principal_client_id,
            service_principal_client_secret=service_principal_client_secret,
            personal_access_token=personal_access_token,
        )


class MosaicSimilarity(Similarity):
    """Mosaic similarity factory.

    Args:
        config: a :class:`MosaicSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("databricks-vectorsearch")

    def ensure_usage_requirements(self):
        fou.ensure_package("databricks-vectorsearch")

    def initialize(self, samples, brain_key):
        return MosaicSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class MosaicSimilarityIndex(SimilarityIndex):
    """Class for interacting with Mosaic similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`MosaicSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`MosaicSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._client = None
        self._index = None
        self._initialize()

    def _initialize(self):
        self._client = vector_search_client.VectorSearchClient(
            workspace_url=self.config.workspace_url,
            service_principal_client_id=self.config.service_principal_client_id,
            service_principal_client_secret=self.config.service_principal_client_secret,
            personal_access_token=self.config.personal_access_token,
        )

        try:
            index_names_result = self._client.list_indexes(
                self.config.endpoint_name
            )
        except Exception as e:
            raise ValueError(
                f"Failed to list indexes from endpoint :{self.config.endpoint_name}"
            ) from e

        index_prefix = f"{self.config.catalog_name}.{self.config.schema_name}."
        if not index_names_result:
            index_names = []
        else:
            index_names = [
                ind["name"].replace(index_prefix, "")
                for ind in index_names_result["vector_indexes"]
                if ind["name"].startswith(index_prefix)
            ]

        if self.config.index_name is None:
            root = "fiftyone-" + fou.to_slug(self._samples._root_dataset.name)
            index_name = fbu.get_unique_name(root, index_names)

            self.config.index_name = index_name
            self.save_config()

        if self.config.index_name in index_names:
            index = self._client.get_index(
                endpoint_name=self.config.endpoint_name,
                index_name=f"{index_prefix}{self.config.index_name}",
            )
        else:
            index = None

        self._index = index

    def _create_index(self, dimension):
        self._index = self._client.create_direct_access_index(
            endpoint_name=self.config.endpoint_name,
            index_name=f"{self.config.catalog_name}.{self.config.schema_name}.{self.config.index_name}",
            primary_key="foid",
            embedding_dimension=dimension,
            embedding_vector_column="embedding_vector",
            schema={
                "foid": "string",
                "sample_id": "string",
                "embedding_vector": "array<float>",
            },
        )

    @property
    def client(self):
        """The ``databricks.vector_search.client.VectorSearchClient`` instance for this index."""
        return self._client

    @property
    def total_index_size(self):
        if self._index is None:
            return 0
        return self._index.describe()["status"]["indexed_row_count"]

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
        batch_size=200,
    ):
        if self._index is None:
            self._create_index(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            index_ids = self._get_index_ids()

            existing_ids = set(ids) & set(index_ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            result = [
                {"foid": f, "sample_id": s, "embedding_vector": list(e)}
                for f, s, e in zip(_ids, _sample_ids, _embeddings)
            ]
            self._index.upsert(result)

        if reload:
            self.reload()

    def _get_index_ids(self, batch_size=200):
        ids = set()
        result = self._index.scan(num_results=batch_size)
        while len(result) > 0:
            ids.update(
                [
                    doc["fields"][0]["value"]["string_value"]
                    for doc in result["data"]
                ]
            )
            last_primary_key = result["last_primary_key"]
            result = self._index.scan(
                num_results=batch_size, last_primary_key=last_primary_key
            )
        return list(ids)

    def _get_values(self, ids, batch_size=200):
        embeddings = []
        result = self._index.scan(num_results=batch_size)
        while len(result) > 0:
            for doc in result["data"]:
                foid = doc["fields"][0]["value"]["string_value"]
                if foid in ids:
                    embedding = [
                        d["number_value"]
                        for d in doc["fields"][2]["value"]["list_value"][
                            "values"
                        ]
                    ]
                    embeddings.append(embedding)
            last_primary_key = result["last_primary_key"]
            result = self._index.scan(
                num_results=batch_size, last_primary_key=last_primary_key
            )

        return embeddings

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = self._get_index_ids()
            missing_ids = set(ids) - set(existing_ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, next(iter(missing_ids)))
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

                ids = existing_ids

        self._index.delete(ids)

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    # Note: might be an arg in delete_brain_run?
    def cleanup(self):
        if self._index is not None:
            self._client.delete_index(
                self.config.endpoint_name,
                f"{self.config.catalog_name}.{self.config.schema_name}.{self.config.index_name}",
            )
            self._index = None

    def _get_sample_embeddings(self, sample_ids, batch_size=200):
        found_embeddings = []
        found_sample_ids = []

        if sample_ids is None:
            sample_ids = self._get_index_ids()

        result = self._index.scan(num_results=batch_size)
        while len(result) > 0:
            for doc in result["data"]:
                sample_id = doc["fields"][1]["value"]["string_value"]
                if sample_id in sample_ids:
                    embedding = [
                        d["number_value"]
                        for d in doc["fields"][2]["value"]["list_value"][
                            "values"
                        ]
                    ]
                    found_embeddings.append(embedding)
                    found_sample_ids.append(sample_id)
            last_primary_key = result["last_primary_key"]
            result = self._index.scan(
                num_results=batch_size, last_primary_key=last_primary_key
            )

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids, batch_size=200):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if label_ids is None:
            label_ids = self._get_index_ids()

        result = self._index.scan(num_results=batch_size)
        while len(result) > 0:
            for doc in result["data"]:
                label_id = doc["fields"][0]["value"]["string_value"]
                if label_id in label_ids:
                    embedding = [
                        d["number_value"]
                        for d in doc["fields"][2]["value"]["list_value"][
                            "values"
                        ]
                    ]
                    found_embeddings.append(embedding)
                    found_label_ids.append(label_id)
                    found_sample_ids.append(
                        doc["fields"][1]["value"]["string_value"]
                    )
            last_primary_key = result["last_primary_key"]
            result = self._index.scan(
                num_results=batch_size, last_primary_key=last_primary_key
            )

        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(
        self, sample_ids, batch_size=200
    ):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        result = self._index.scan(num_results=batch_size)
        while len(result) > 0:
            for doc in result["data"]:
                sample_id = doc["fields"][1]["value"]["string_value"]
                if sample_id in sample_ids:
                    embedding = [
                        d["number_value"]
                        for d in doc["fields"][2]["value"]["list_value"][
                            "values"
                        ]
                    ]
                    found_embeddings.append(embedding)
                    found_sample_ids.append(sample_id)
                    found_label_ids.append(
                        doc["fields"][0]["value"]["string_value"]
                    )
            last_primary_key = result["last_primary_key"]
            result = self._index.scan(
                num_results=batch_size, last_primary_key=last_primary_key
            )

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):

        if query is None:
            raise ValueError("Mosaic does not support full index neighbors")

        if reverse is True:
            raise ValueError(
                "Mosaic does not support least similarity queries"
            )

        if k is None:
            k = self.index_size

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = list(self.current_label_ids)
            else:
                index_ids = list(self.current_sample_ids)

            _filter = {"foid": list(index_ids)}
            # NOTE: Filtering is supported in Mosaic but is not robust and cannot handle a large number of filters
            # so we apply the filter after the search
        else:
            _filter = None

        ids = []
        dists = []
        for q in query:
            results = self._index.similarity_search(
                columns=["foid"],
                query_vector=[float(i) for i in list(q)],
                num_results=k,
            )

            if _filter:
                ids.append(
                    [
                        res[0]
                        for res in results["result"]["data_array"]
                        if res[0] in _filter["foid"]
                    ]
                )
            else:
                ids.append([res[0] for res in results["result"]["data_array"]])
            if return_dists:
                dists.append(
                    [res[1] for res in results["result"]["data_array"]]
                )

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        embeddings = self._get_values(query_ids)
        if len(embeddings) == 0:
            raise ValueError(
                "Query IDs %s do not exist in this index" % query_ids
            )
        query = np.array(embeddings)

        if single_query:
            query = query[0, :]

        return query

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
Redis similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

redis = fou.lazy_import("redis")


logger = logging.getLogger(__name__)

_SUPPORTED_METRICS = {
    "cosine": "COSINE",
    "dotproduct": "IP",
    "euclidean": "L2",
}


class RedisSimilarityConfig(SimilarityConfig):
    """Configuration for the Redis similarity backend.

    Args:
        index_name (None): the name of a Redis index to use or create. If none
            is provided, a new index will be created
        metric ("cosine"): the embedding distance metric to use when creating a
            new index. Supported values are
            ``("cosine", "dotproduct", "euclidean")``
        algorithm ("FLAT"): the search algorithm to use. The supported values
            are ``("FLAT", "HNSW")``
        host ("localhost"): the host to use
        port (6379): the port to use
        db (0): the database to use
        username (None): a username to use
        password (None): a password to use
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        index_name=None,
        metric="cosine",
        algorithm="FLAT",
        host="localhost",
        port=6379,
        db=0,
        username=None,
        password=None,
        **kwargs,
    ):
        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                "Unsupported metric '%s'. Supported values are %s"
                % (metric, tuple(_SUPPORTED_METRICS.keys()))
            )

        super().__init__(**kwargs)

        self.index_name = index_name
        self.metric = metric
        self.algorithm = algorithm

        # store privately so these aren't serialized
        self._host = host
        self._port = port
        self._db = db
        self._username = username
        self._password = password

    @property
    def method(self):
        return "redis"

    @property
    def host(self):
        return self._host

    @host.setter
    def host(self, value):
        self._host = value

    @property
    def port(self):
        return self._port

    @port.setter
    def port(self, value):
        self._port = value

    @property
    def db(self):
        return self._db

    @db.setter
    def db(self, value):
        self._db = value

    @property
    def username(self):
        return self._username

    @username.setter
    def username(self, value):
        self._username = value

    @property
    def password(self):
        return self._password

    @password.setter
    def password(self, value):
        self._password = value

    @property
    def max_k(self):
        return None

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self, host=None, port=None, db=None, username=None, password=None
    ):
        self._load_parameters(
            host=host, port=port, db=db, username=username, password=password
        )


class RedisSimilarity(Similarity):
    """Redis similarity factory.

    Args:
        config: a :class:`RedisSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("redis")

    def ensure_usage_requirements(self):
        fou.ensure_package("redis")

    def initialize(self, samples, brain_key):
        return RedisSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class RedisSimilarityIndex(SimilarityIndex):
    """Class for interacting with Redis similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`RedisSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`RedisSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._client = None
        self._index = None
        self._initialize()

    def _initialize(self):
        client = redis.Redis(
            host=self.config.host,
            port=self.config.port,
            db=self.config.db,
            username=self.config.username,
            password=self.config.password,
            decode_responses=True,
        )

        if self.config.index_name is None:

            def index_exists(index_name):
                try:
                    client.ft(index_name).info()
                    return True
                except:
                    return False

            root = "fiftyone-" + fou.to_slug(self._samples._root_dataset.name)
            index_name = fbu.get_unique_name(root, index_exists)

            self.config.index_name = index_name
            self.save_config()

        try:
            index = client.ft(self.config.index_name)
            index.info()
        except:
            index = None

        self._client = client
        self._index = index

    def _create_index(self, dimension):
        from redis.commands.search.field import TagField, VectorField
        from redis.commands.search.indexDefinition import (
            IndexDefinition,
            IndexType,
        )

        schema = (
            TagField("$.foid", as_name="foid"),
            TagField("$.sample_id", as_name="sample_id"),
            VectorField(
                "$.vector",
                self.config.algorithm,
                {
                    "TYPE": "FLOAT32",
                    "DIM": dimension,
                    "DISTANCE_METRIC": _SUPPORTED_METRICS[self.config.metric],
                },
                as_name="vector",
            ),
        )
        definition = IndexDefinition(
            prefix=[self.config.index_name + ":"],
            index_type=IndexType.JSON,
        )
        index = self._client.ft(self.config.index_name)
        index.create_index(fields=schema, definition=definition)

        self._index = index

    @property
    def client(self):
        """The ``redis.client.Redis`` instance for this index."""
        return self._client

    @property
    def total_index_size(self):
        try:
            return int(self._index.info()["num_docs"])
        except:
            return 0

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
    ):
        if self._index is None:
            self._create_index(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            existing_ids = self._get_existing_ids(ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            del_inds = [i for i, _id in enumerate(ids) if _id in existing_ids]
            embeddings = np.delete(embeddings, del_inds, axis=0)
            sample_ids = np.delete(sample_ids, del_inds)
            if label_ids is not None:
                label_ids = np.delete(label_ids, del_inds)
        elif existing_ids and overwrite:
            self._delete_ids(existing_ids)

        pipeline = self._client.pipeline()
        for e, id, sample_id in zip(embeddings, ids, sample_ids):
            key = f"{self.config.index_name}:{id}"
            d = {
                "foid": id,
                "sample_id": sample_id,
                "vector": e.astype(np.float32).tolist(),
            }
            pipeline.json().set(key, "$", d)

        pipeline.execute()

        if reload:
            self.reload()

    def _get_existing_ids(self, ids):
        return [d["foid"] for d in self._get_values(ids)]

    def _delete_ids(self, ids):
        keys = [f"{self.config.index_name}:{id}" for id in ids]
        self._client.delete(*keys)

    def _get_values(self, ids):
        pipeline = self._client.pipeline()
        for id in ids:
            pipeline.json().get(f"{self.config.index_name}:{id}")

        return [d for d in pipeline.execute() if d is not None]

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if not allow_missing or warn_missing:
            existing_ids = self._get_existing_ids(ids)
            missing_ids = set(ids) - set(existing_ids)
            num_missing = len(missing_ids)

            if num_missing > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that are not present in the "
                        "index" % (num_missing, next(iter(missing_ids)))
                    )

                if warn_missing:
                    logger.warning(
                        "Ignoring %d IDs that are not present in the index",
                        num_missing,
                    )

                ids = existing_ids

        self._delete_ids(ids=ids)

        if reload:
            self.reload()

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_sample_ids(sample_ids)
        elif self.config.patches_field is not None:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_patch_embeddings_from_label_ids(label_ids)
        else:
            (
                embeddings,
                sample_ids,
                label_ids,
                missing_ids,
            ) = self._get_sample_embeddings(sample_ids)

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def cleanup(self):
        if self._index is None:
            return

        self._index.dropindex(delete_documents=True)
        self._index = None

    def _get_sample_embeddings(self, sample_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []

        if sample_ids is None:
            get_id = lambda key: key.rsplit(":", 1)[1]
            keys = self._client.keys(f"{self.config.index_name}:*")
            sample_ids = map(get_id, keys)

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            for d in self._get_values(batch_ids):
                found_embeddings.append(d["vector"])
                found_sample_ids.append(d["sample_id"])

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, None, missing_ids

    def _get_patch_embeddings_from_label_ids(self, label_ids, batch_size=1000):
        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        if label_ids is None:
            get_id = lambda key: key.rsplit(":", 1)[1]
            keys = self._client.keys(f"{self.config.index_name}:*")
            label_ids = map(get_id, keys)

        for batch_ids in fou.iter_batches(label_ids, batch_size):
            for d in self._get_values(batch_ids):
                found_embeddings.append(d["vector"])
                found_sample_ids.append(d["sample_id"])
                found_label_ids.append(d["foid"])

        missing_ids = list(set(label_ids) - set(found_label_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _get_patch_embeddings_from_sample_ids(
        self, sample_ids, batch_size=100
    ):
        from redis.commands.search.query import Query

        found_embeddings = []
        found_sample_ids = []
        found_label_ids = []

        for batch_ids in fou.iter_batches(sample_ids, batch_size):
            filter = "@sample_id:{ " + " | ".join(batch_ids) + " }"
            query = Query(filter).dialect(2)
            for doc in self._index.search(query).docs:
                found_embeddings.append(doc.embeddings)
                found_sample_ids.append(doc.sample_id)
                found_label_ids.append(doc.foid)

        missing_ids = list(set(sample_ids) - set(found_sample_ids))

        return found_embeddings, found_sample_ids, found_label_ids, missing_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
    ):
        from redis.commands.search.query import Query

        if query is None:
            raise ValueError("Redis does not support full index neighbors")

        if reverse is True:
            raise ValueError("Redis does not support least similarity queries")

        if k is None:
            k = self.index_size

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = list(self.current_label_ids)
            else:
                index_ids = list(self.current_sample_ids)

            filter = "@foid:{ " + " | ".join(index_ids) + " }"
        else:
            filter = "*"

        ids = []
        dists = []
        for q in query:
            _query = (
                Query(f"({filter})=>[KNN {k} @vector $query AS score]")
                .sort_by("score")
                .return_fields("score", "foid")
                .dialect(2)
                .paging(0, k)
            )
            _q = q.astype(np.float32).tobytes()
            docs = self._index.search(_query, {"query": _q}).docs

            ids.append([doc.foid for doc in docs])
            if return_dists:
                dists.append([doc.score for doc in docs])

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        # Query by ID(s)
        dicts = self._get_values(query_ids)
        if not dicts:
            raise ValueError(
                "Query IDs %s do not exist in this index" % query_ids
            )

        query = np.array([d["vector"] for d in dicts])

        if single_query:
            query = query[0, :]

        return query

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)



"""
PGVector similarity backend.

| Copyright 2017-2025, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
import logging

import numpy as np

import eta.core.utils as etau

import fiftyone.core.utils as fou
from fiftyone.brain.similarity import (
    SimilarityConfig,
    Similarity,
    SimilarityIndex,
)
import fiftyone.brain.internal.core.utils as fbu

psycopg2 = fou.lazy_import("psycopg2")
psy_extras = fou.lazy_import("psycopg2.extras")

logger = logging.getLogger(__name__)

# Supported metrics for pgvector
_SUPPORTED_METRICS = {
    "cosine": "vector_cosine_ops",
    "dotproduct": "vector_ip_ops",
    "euclidean": "vector_l2_ops",
    "l1": "vector_l1_ops",
    "jaccard": "vector_jaccard_ops",
    "hamming": "vector_hamming_ops",
}


class PgVectorSimilarityConfig(SimilarityConfig):
    """Configuration for the PGVector similarity backend.

    Args:
        index_name (None): the name of the PGVector index to use or create.
            If none is provided, a default index name will be used.
        table_name (None): the name of the table to use or create. If none is
            provided, a default table name will be used.
        metric ("cosine"): the similarity metric to use. Supported values are
            ``("cosine", "dotproduct", "euclidean", "l1", "jaccard", "hamming")``
        connection_string (None): the connection string to the PostgreSQL database
        ssl_cert (None): the path to the SSL certificate file
        ssl_key (None): the path to the secret key used for the client certificate
        ssl_root_cert (None): the path to the file containing SSL certificate
            authority (CA) certificate(s).
        work_mem ("64MB"): the base maximum amount of memory to be used by a query operation
            (such as a sort or hash table) before writing to temporary disk files
        hnsw_m (16): the max number of connections per layer in the HNSW index
        hnsw_ef_construction (64): the size of the dynamic candidate list for constructing the graph for the HNSW index
        **kwargs: keyword arguments for
            :class:`fiftyone.brain.similarity.SimilarityConfig`
    """

    def __init__(
        self,
        index_name=None,
        table_name=None,
        metric="cosine",
        connection_string=None,
        ssl_cert=None,
        ssl_key=None,
        ssl_root_cert=None,
        work_mem="64MB",
        hnsw_m=16,
        hnsw_ef_construction=64,
        **kwargs,
    ):
        if metric not in _SUPPORTED_METRICS:
            raise ValueError(
                f"Unsupported metric '{metric}'. "
                f"Supported values are {_SUPPORTED_METRICS}"
            )

        super().__init__(**kwargs)

        self.metric = metric
        self.ssl_cert = ssl_cert
        self.ssl_key = ssl_key
        self.ssl_root_cert = ssl_root_cert
        self.work_mem = work_mem
        self.index_name = index_name
        self.table_name = table_name
        self.hnsw_m = hnsw_m
        self.hnsw_ef_construction = hnsw_ef_construction

        self._connection_string = connection_string

    @property
    def method(self):
        return "pgvector"

    @property
    def connection_string(self):
        return self._connection_string

    @connection_string.setter
    def connection_string(self, connection_string):
        self._connection_string = connection_string

    @property
    def max_k(self):
        return 10000

    @property
    def supports_least_similarity(self):
        return False

    @property
    def supported_aggregations(self):
        return ("mean",)

    def load_credentials(
        self,
        connection_string=None,
    ):
        self._load_parameters(connection_string=connection_string)


class PgVectorSimilarity(Similarity):
    """PGVector similarity factory.

    Args:
        config: a :class:`PgVectorSimilarityConfig`
    """

    def ensure_requirements(self):
        fou.ensure_package("psycopg2-binary")

    def ensure_usage_requirements(self):
        fou.ensure_package("psycopg2-binary")

    def initialize(self, samples, brain_key):
        return PgVectorSimilarityIndex(
            samples, self.config, brain_key, backend=self
        )


class PgVectorSimilarityIndex(SimilarityIndex):
    """Class for interacting with PGVector similarity indexes.

    Args:
        samples: the :class:`fiftyone.core.collections.SampleCollection` used
        config: the :class:`PGVectorSimilarityConfig` used
        brain_key: the brain key
        backend (None): a :class:`PGVectorSimilarity` instance
    """

    def __init__(self, samples, config, brain_key, backend=None):
        super().__init__(samples, config, brain_key, backend=backend)
        self._conn = None
        self._cur = None
        self._initialize()

    @property
    def total_index_size(self):
        if self._conn.closed:
            self._initialize()
        try:
            self._cur.execute(
                f"""SELECT COUNT(*) FROM "{self.config.table_name}";"""
            )
            return self._cur.fetchone()[0]
        except Exception as e:
            logger.error(f"Error getting index size: {str(e)}")
            return 0

    def _initialize(self):
        ssl_options = {}
        if self.config.ssl_cert:
            ssl_options["sslcert"] = self.config.ssl_cert
        if self.config.ssl_key:
            ssl_options["sslkey"] = self.config.ssl_key
        if self.config.ssl_root_cert:
            ssl_options["sslrootcert"] = self.config.ssl_root_cert

        logger.info(f"Connecting to PostgreSQL database")
        self._conn = psycopg2.connect(
            self.config.connection_string, **ssl_options
        )
        self._cur = self._conn.cursor()
        try:
            self._cur.execute("CREATE EXTENSION IF NOT EXISTS vector")
            self._conn.commit()
        except Exception as e:
            logger.error(f"Error creating vector extension: {str(e)}")
            raise

        if self.config.table_name is None:
            table_names = self._get_table_names()
            root = "fiftyone-" + fou.to_slug(self.samples._root_dataset.name)
            table_name = fbu.get_unique_name(root, table_names)

            self.config.table_name = table_name
            self.save_config()
            existing_indexes = []
        else:
            existing_indexes = self._get_index_names(self.config.table_name)

        if self.config.index_name is None:
            root = "fiftyone-index-" + fou.to_slug(
                self.samples._root_dataset.name
            )
            index_name = fbu.get_unique_name(root, existing_indexes)
            self.config.index_name = index_name
            self.save_config()

    def _get_table_names(self):
        self._cur.execute(
            "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public';"
        )
        return [row[0] for row in self._cur.fetchall()]

    def _get_index_names(self, table_name):
        self._cur.execute(
            f"SELECT indexname FROM pg_indexes WHERE tablename = '{table_name}' AND schemaname = 'public';"
        )
        return [row[0] for row in self._cur.fetchall()]

    def _create_table(self, dimension):
        try:
            self._cur.execute(
                f"""
                    CREATE TABLE IF NOT EXISTS "{self.config.table_name}" (
                    id TEXT PRIMARY KEY,
                    sample_id TEXT,
                    embedding_vector VECTOR({dimension})
                );
                """
            )
            self._conn.commit()
        except Exception as e:
            logger.error(
                f"Error creating table: {self.config.table_name} with dimension {dimension}: {str(e)}"
            )
            raise

    def create_hnsw_index(self):
        operator_class = _SUPPORTED_METRICS[self.config.metric]
        try:
            self._cur.execute(
                f"""DROP INDEX IF EXISTS "{self.config.index_name}";"""
            )
            self._conn.commit()
            self._cur.execute(
                f"""
                CREATE INDEX "{self.config.index_name}"
                ON "{self.config.table_name}" USING hnsw (embedding_vector {operator_class})
                WITH (m = %s, ef_construction = %s);
                """,
                (self.config.hnsw_m, self.config.hnsw_ef_construction),
            )
            self._conn.commit()
        except Exception as e:
            logger.error(
                f"Error creating HNSW index on table {self.config.table_name}:{str(e)}"
            )
            raise

    def _get_index_ids(self, batch_size=1000):
        named_cursor = self._conn.cursor(
            name="id_cursor"
        )  # Named cursor for server-side query
        named_cursor.execute(f"""SELECT id FROM "{self.config.table_name}";""")

        existing_ids = []
        while True:
            rows = named_cursor.fetchmany(batch_size)
            if not rows:
                break

            ids = [row[0] for row in rows]
            existing_ids.extend(ids)

        named_cursor.close()
        return existing_ids

    def add_to_index(
        self,
        embeddings,
        sample_ids,
        label_ids=None,
        overwrite=True,
        allow_existing=True,
        warn_existing=False,
        reload=True,
        batch_size=5000,
        close_conn=True,
    ):
        if self._conn.closed:
            self._initialize()
        self._cur.execute(f"SET work_mem TO '{self.config.work_mem}'")

        if self.config.table_name not in self._get_table_names():
            self._create_table(embeddings.shape[1])

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_existing or not allow_existing or not overwrite:
            index_ids = self._get_index_ids()

            existing_ids = set(ids) & set(index_ids)
            num_existing = len(existing_ids)

            if num_existing > 0:
                if not allow_existing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that already exist in the index"
                        % (num_existing, next(iter(existing_ids)))
                    )

                if warn_existing:
                    if overwrite:
                        logger.warning(
                            "Overwriting %d IDs that already exist in the "
                            "index",
                            num_existing,
                        )
                    else:
                        logger.warning(
                            "Skipping %d IDs that already exist in the index",
                            num_existing,
                        )
        else:
            existing_ids = set()

        if existing_ids and not overwrite:
            query = f"""
                INSERT INTO "{self.config.table_name}" (id, sample_id, embedding_vector)
                VALUES %s
                ON CONFLICT (id) DO NOTHING;
                """
        else:
            query = f"""
                INSERT INTO "{self.config.table_name}" (id, sample_id, embedding_vector)
                VALUES %s
                ON CONFLICT (id) DO UPDATE
                SET sample_id = EXCLUDED.sample_id,
                    embedding_vector = EXCLUDED.embedding_vector;
                """

        embeddings = [e.tolist() for e in embeddings]
        sample_ids = list(sample_ids)
        if label_ids is not None:
            ids = list(label_ids)
        else:
            ids = list(sample_ids)

        for _embeddings, _ids, _sample_ids in zip(
            fou.iter_batches(embeddings, batch_size),
            fou.iter_batches(ids, batch_size),
            fou.iter_batches(sample_ids, batch_size),
        ):
            data = list(zip(_ids, _sample_ids, _embeddings))
            psy_extras.execute_values(self._cur, query, data)
            self._conn.commit()

        self.create_hnsw_index()

        if close_conn:
            self.close_connections()

        if reload:
            self.reload()

    def remove_from_index(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
        reload=True,
    ):
        if self._conn.closed:
            self._initialize()

        if label_ids is not None:
            ids = label_ids
        else:
            ids = sample_ids

        if warn_missing or not allow_missing:
            response = self.get_embeddings_by_id(ids)
            existing_ids = [id for id, emb in response]
            missing_ids = set(ids) - set(existing_ids)
            num_missing_ids = len(missing_ids)

            if num_missing_ids > 0:
                if not allow_missing:
                    raise ValueError(
                        "Found %d IDs (eg %s) that do not exist in the index"
                        % (num_missing_ids, next(iter(missing_ids)))
                    )
                if warn_missing and not allow_missing:
                    logger.warning(
                        "Skipping %d IDs that do not exist in the index",
                        num_missing_ids,
                    )
        try:
            # Use parameterized query to delete multiple IDs
            self._cur.execute(
                f"""DELETE FROM "{self.config.table_name}" WHERE id IN %s;""",
                (tuple(ids),),
            )
        except Exception as e:
            self._conn.rollback()
            logger.error(f"Error removing embeddings for ids {ids}: {str(e)}")
            raise

        deleted_count = self._cur.rowcount
        self._conn.commit()
        logger.info(f"Deleted {deleted_count} embeddings from the index.")

        if reload:
            self.reload()

    def close_connections(self):
        if not self._cur.closed:
            self._cur.close()
        if not self._conn.closed:
            self._conn.close()

    def get_embeddings_by_id(self, sample_ids=None, label_ids=None):
        if self._conn.closed:
            self._initialize()
        if label_ids is not None:
            try:
                self._cur.execute(
                    f"""SELECT id, sample_id, embedding_vector FROM "{self.config.table_name}" WHERE id = ANY(%s)""",
                    (list(label_ids),),
                )
            except Exception as e:
                logger.error(
                    f"Error fetching embeddings for labels {label_ids}: {str(e)}"
                )
                raise
        elif sample_ids is not None:
            try:
                self._cur.execute(
                    f"""SELECT id, sample_id, embedding_vector FROM "{self.config.table_name}" WHERE sample_id = ANY(%s)""",
                    (list(sample_ids),),
                )
            except Exception as e:
                logger.error(
                    f"Error fetching embeddings for samples {sample_ids}: {str(e)}"
                )
                raise
        else:
            try:
                self._cur.execute(
                    f"""SELECT id, sample_id, embedding_vector FROM "{self.config.table_name}";"""
                )
            except Exception as e:
                logger.error(
                    f"Error fetching embeddings for all samples: {str(e)}"
                )
                raise

        results = self._cur.fetchall()
        fo_id = []
        sample_id = []
        embeddings = []
        for result in results:
            # Convert string "[1.2,3.4,5.6]" to float array
            if isinstance(result[2], str):
                emb = np.array(
                    [float(x) for x in result[2].strip("[]").split(",")],
                    dtype=np.float32,
                )
                embeddings.append(emb)
            else:
                # Already numeric
                emb = np.array(result[2], dtype=np.float32)
                embeddings.append(emb)
            fo_id.append(result[0])
            sample_id.append(result[1])

        return fo_id, sample_id, embeddings

    def get_embeddings(
        self,
        sample_ids=None,
        label_ids=None,
        allow_missing=True,
        warn_missing=False,
    ):
        if label_ids is not None:
            if self.config.patches_field is None:
                raise ValueError("This index does not support label IDs")

            if sample_ids is not None:
                logger.warning(
                    "Ignoring sample IDs when label IDs are provided"
                )

        if sample_ids is not None and self.config.patches_field is not None:
            (
                label_ids,
                found_sample_ids,
                embeddings,
            ) = self.get_embeddings_by_id(sample_ids=sample_ids)
            missing_ids = list(set(sample_ids) - set(found_sample_ids))
            sample_ids = found_sample_ids
        elif self.config.patches_field is not None:
            (
                found_label_ids,
                sample_ids,
                embeddings,
            ) = self.get_embeddings_by_id(label_ids=label_ids)
            missing_ids = (
                list(set(label_ids) - set(found_label_ids))
                if label_ids is not None
                else []
            )
            label_ids = found_label_ids
        else:
            (
                label_ids,
                found_sample_ids,
                embeddings,
            ) = self.get_embeddings_by_id(sample_ids=sample_ids)
            missing_ids = (
                list(set(sample_ids) - set(found_sample_ids))
                if sample_ids is not None
                else []
            )
            sample_ids = found_sample_ids

        num_missing_ids = len(missing_ids)
        if num_missing_ids > 0:
            if not allow_missing:
                raise ValueError(
                    "Found %d IDs (eg %s) that do not exist in the index"
                    % (num_missing_ids, missing_ids[0])
                )

            if warn_missing:
                logger.warning(
                    "Skipping %d IDs that do not exist in the index",
                    num_missing_ids,
                )

        embeddings = np.array(embeddings)
        sample_ids = np.array(sample_ids)
        if label_ids is not None:
            label_ids = np.array(label_ids)

        return embeddings, sample_ids, label_ids

    def _kneighbors(
        self,
        query=None,
        k=None,
        reverse=False,
        aggregation=None,
        return_dists=False,
        close_conn=True,
    ):
        if self._conn.closed:
            self._initialize()

        if query is None:
            raise ValueError("Postgres does not support full index neighbors")

        if aggregation not in (None, "mean"):
            raise ValueError("Unsupported aggregation '%s'" % aggregation)

        if k is None:
            k = self.index_size

        query = self._parse_neighbors_query(query)
        if aggregation == "mean" and query.ndim == 2:
            query = query.mean(axis=0)

        single_query = query.ndim == 1
        if single_query:
            query = [query]

        index_ids = None
        if self.has_view:
            if self.config.patches_field is not None:
                index_ids = self.current_label_ids
            else:
                index_ids = self.current_sample_ids

            _filter = True
        else:
            _filter = False

        sort_order = "DESC" if reverse else "ASC"

        ids = []
        dists = []
        for q in query:
            q = q.tolist()
            if _filter:
                self._cur.execute(
                    f"""
                    SELECT id, embedding_vector <-> %s::vector AS distance
                    FROM "{self.config.table_name}"
                    WHERE id = ANY(%s)
                    ORDER BY distance {sort_order}
                    LIMIT %s;
                    """,
                    (q, list(index_ids), k),
                )
            else:
                self._cur.execute(
                    f"""
                    SELECT id, embedding_vector <-> %s::vector AS distance
                    FROM "{self.config.table_name}"
                    ORDER BY distance {sort_order}
                    LIMIT %s;
                    """,
                    (q, k),
                )

            results = self._cur.fetchall()
            ids.append([r[0] for r in results])
            distances = [r[1] for r in results]

            if return_dists:
                dists.append(distances)

        if single_query:
            ids = ids[0]
            if return_dists:
                dists = dists[0]

        if close_conn:
            self.close_connections()

        if return_dists:
            return ids, dists

        return ids

    def _parse_neighbors_query(self, query):
        if etau.is_str(query):
            query_ids = [query]
            single_query = True
        else:
            query = np.asarray(query)

            # Query by vector(s)
            if np.issubdtype(query.dtype, np.number):
                return query

            query_ids = list(query)
            single_query = False

        _, _, embeddings = self.get_embeddings_by_id(label_ids=query_ids)
        if len(embeddings) == 0:
            raise ValueError(
                "Query IDs %s do not exist in this index" % query_ids
            )
        query = np.array(embeddings)

        if single_query:
            query = query[0, :]

        return query

    def cleanup(self, drop_table=False):
        """
        Clean up the database by dropping the HNSW index and optionally the embeddings table.
        """
        logger.info(
            f"Cleaning up: Deleting HNSW index '{self.config.index_name}'"
        )
        self._cur.execute(
            f"""DROP INDEX IF EXISTS "{self.config.index_name}";"""
        )

        if self._conn.closed:
            self._initialize()

        if drop_table:
            self._cur.execute(
                f"""DROP TABLE IF EXISTS "{self.config.table_name}";"""
            )
            logger.info(
                f"{self.config.table_name} table deleted successfully."
            )

        self._conn.commit()
        # Close the database connection
        self.close_connections()
        logger.info("Database connection closed.")

    @classmethod
    def _from_dict(cls, d, samples, config, brain_key):
        return cls(samples, config, brain_key)
